<?php
/**
 * +----------------------------------------------------------------------
 * 微服务-客户端调用类
 * +----------------------------------------------------------------------
 * 官网：https://www.sw-x.cn
 * +----------------------------------------------------------------------
 * 作者：小黄牛 <1731223728@qq.com>
 * +----------------------------------------------------------------------
 * 开源协议：http://www.apache.org/licenses/LICENSE-2.0
 * +----------------------------------------------------------------------
*/

namespace SwooleXRpc;

class RpcClient
{   
    /**
     * 请求结果 true|false
    */
    private $status = false;
    /**
     * 返回说明
    */
    private $msg = 'ERROR';
    /**
     * 返回结果
    */
    private $data;
    /**
     * 请求开始时间
    */
    private $start_time;
    /**
     * 请求超时时间(S)
    */
    private $out_time;
    /**
     * 发送次数
    */
    private $send_num = 1;
    /**
     * 请求路由
    */
    private $class;
    /**
     * 请求方法
    */
    private $function;
    /**
     * 是否异步执行
    */
    private $task=false;
    /**
     * 请求头
    */
    private $headers = array();
    /**
     * 请求参数
    */
    private $param = array();
    /**
     * 失败时，使用最大节点次数
    */
    private $num = 1;
    /**
     * 开始请求时间
    */
    private $start_ms;
    /**
     * 当前请求配置
    */
    private $config;
    /**
     * RPC客户端配置
    */
    private $rpc;
    /**
     * Redis配置
    */
    private $redis;
    /**
     * 异步回调地址
    */
    private $callback = false;
    /**
     * 异步回调请求类型
    */
    private $callback_type = 'post';

    /**
     * 标记开始时间
     * @todo 无
     * @author 小黄牛
     * @version v1.2.24 + 2021.1.9
     * @deprecated 暂不启用
     * @global 无
     * @return void
    */
    public function __construct() {
        $this->start_time = time();
        $this->start_ms = microtime(true);
        $this->rpc = require dirname(__FILE__).'/config/rpc.php';
        $this->out_time = $this->rpc['out_time'];
        $this->redis = require dirname(__FILE__).'/config/rpc.php';
    }

    /**
     * 判断单个请求延迟
     * @todo 无
     * @author 小黄牛
     * @version v1.2.24 + 2021.1.9
     * @deprecated 暂不启用
     * @global 无
     * @return void
    */
    public function __destruct() {
        if (!empty($this->config['max_ms'])) {
            $end_ms = microtime(true);
            $ms = ($end_ms-$this->start_ms)*1000;
            if ($ms >= $this->config['max_ms']) {
                $msg  = '请求耗时（ms）：'.$ms.PHP_EOL;
                $msg .= '请求路由：'.$this->class.PHP_EOL;
                $msg .= '请求方法：'.$this->function.PHP_EOL;
                $msg .= '请求头：'.json_encode($this->headers).PHP_EOL;
                $msg .= '请求参数：'.json_encode($this->param).PHP_EOL;
                $msg .= '请求节点：'.json_encode($this->config).PHP_EOL.PHP_EOL;

                $dir = dirname(__FILE__).'/runtime/rpc/';
                if (is_dir($dir) == false) {
                    mkdir($dir, 0755);
                }

                $file_path = $dir.date('Ymd').'.log';
                // 写入日志记录
                file_put_contents($file_path, $msg, FILE_APPEND);
            }
        }
    }

    /**
     * 设置参数
     * @todo 无
     * @author 小黄牛
     * @version v1.2.24 + 2021.1.9
     * @deprecated 暂不启用
     * @global 无
     * @param string $key
     * @param mixed $val
     * @return void
    */
    public function set($key, $val) {
        $this->$key = $val;
        return $this;
    }

    /**
     * 设置路由地址
     * @todo 无
     * @author 小黄牛
     * @version v1.2.24 + 2021.1.9
     * @deprecated 暂不启用
     * @global 无
     * @param string $class 请求路由
     * @return void
    */
    public function route($class) {
        $this->class = $class;
        return $this;
    }

    /**
     * 设置请求方法
     * @todo 无
     * @author 小黄牛
     * @version v1.2.24 + 2021.1.9
     * @deprecated 暂不启用
     * @global 无
     * @param string $function 请求方法
     * @return void
    */
    public function func($function) {
        $this->function = $function;
        return $this;
    }

    /**
     * 设置为异步任务
     * @todo 无
     * @author 小黄牛
     * @version v1.2.24 + 2021.1.9
     * @deprecated 暂不启用
     * @global 无
     * @return void
    */
    public function task() {
        $this->task = true;
        return $this;
    }
    
    /**
     * 设置请求头
     * @todo 无
     * @author 小黄牛
     * @version v1.2.24 + 2021.1.9
     * @deprecated 暂不启用
     * @global 无
     * @param array $header 请求头
     * @return void
    */
    public function header($header) {
        $this->headers = $header;
        return $this;
    }

    /**
     * 设置请求参数
     * @todo 无
     * @author 小黄牛
     * @version v1.2.24 + 2021.1.9
     * @deprecated 暂不启用
     * @global 无
     * @param array $param 参数
     * @return void
    */
    public function param($param) {
        $this->param = $param;
        return $this;
    }

    /**
     * 设置最大请求次数
     * @todo 无
     * @author 小黄牛
     * @version v1.2.24 + 2021.1.9
     * @deprecated 暂不启用
     * @global 无
     * @param string $num 次数
     * @return void
    */
    public function max($num) {
        $this->num = $num;
        return $this;
    }

    /**
     * 设置异步任务的回调通知地址
     * @todo 无
     * @author 小黄牛
     * @version v1.2.24 + 2021.1.9
     * @deprecated 暂不启用
     * @global 无
     * @param string $url
     * @param string $type
     * @return void
    */
    public function callback($url, $type='post') {
        $this->callback = $url;
        $this->callback_type = strtolower($type);
        return $this;
    }

    /**
     * 请求微服务
     * @todo 无
     * @author 小黄牛
     * @version v1.2.24 + 2021.1.9
     * @deprecated 暂不启用
     * @global 无
     * @return mixed
    */
    public function send() {
        if ((time()-$this->start_time) >= $this->out_time) {
            $this->msg = "rpc request timeout";
            $this->status = '408';
            return false;
        }

        $list = \SwooleXRpc\Rpc::run()->get($this->class, $this->function);
        if (empty($list)) {
            $this->msg = "rpc service 【".$this->class." ".$this->function."】 not registered";
            return false;
        }

        // 递归到最后一个节点了
        if ($this->num < $this->send_num) {
            return false;
        }
        $config = $this->weightConfig($list);
        if ($config == false) {
            $this->msg = "rpc The service has been completely stopped";
            \SwooleXRpc\Rpc::run()->ping_error(['class'=>$this->class, 'function'=>$this->function], 3);
            return false;
        }

        // 发送请求
        $res = $this->run($config);
        if ($this->status != '200') {
            return $this->send();
        }

        return $res;
    }

    // 发送微服务请求
    private function run($config) {
        $this->config = $config;
        $this->send_num++;

        // 更新当前请求数
        $md5 = md5($config['class'].$config['function'].$config['ip'].$config['port']);
        $num_key = $this->redis['redis_key'].'_num_'.$md5;
        \SwooleXRpc\RpcRedis::run()->INCR($num_key); 

        $data = json_encode(array(
            'class' => $this->class,
            'function' => $this->function,
            'headers' => $this->headers,
            'ip' => $config['ip'],
            'port' => $config['port'],
            'param' => $this->param,
            'task' => $this->task,
            'callback' => $this->callback,
            'callback_type' => $this->callback_type,
        ), JSON_UNESCAPED_UNICODE);

        // 数据加密
        if ($this->rpc['aes_status'] == true) {
            $Currency = new \SwooleXRpc\x\Currency();
            $data = $Currency->aes_encrypt($data, $this->rpc);
            unset($Currency);
        }

        // 调用服务
        try {
            $length = strlen($data);
            //创建tcp套接字
            $socket = socket_create(AF_INET,SOCK_STREAM,SOL_TCP);
            //连接超时
            socket_set_option($socket,SOL_SOCKET,SO_RCVTIMEO,array("sec"=>$this->rpc['out_time'], "usec"=>0 ) );
            //接收超时
            socket_set_option($socket,SOL_SOCKET,SO_SNDTIMEO,array("sec"=>$this->rpc['write_timeout'], "usec"=>0 ) );
            //连接tcp
            $res = @socket_connect($socket, $config['ip'], $config['port']);
            if (!$res) {
                $this->msg = 'connect failed. Error: connection failed！';
                \SwooleXRpc\RpcRedis::run()->DECR($num_key); 
                // 这里理应关闭该连接，标记is_fault
                $config['is_fault'] = 1;
                \SwooleXRpc\Rpc::run()->set($config);
                \SwooleXRpc\Rpc::run()->ping_error($config, 4);
                //关闭连接
                socket_close($socket);
                return false;
            }
            //向打开的套集字写入数据（发送数据）
            $s = socket_write($socket, $data, $length);
            //从套接字中获取服务器发送来的数据
            $body = socket_read($socket, $this->rpc['string_max_size']);
            //关闭连接
            socket_close($socket);
        } catch (\Exception $e) {
            $this->msg = 'connect failed. Error: '.$e->getMessage();
            \SwooleXRpc\RpcRedis::run()->DECR($num_key); 
            // 这里理应关闭该连接，标记is_fault
            $config['is_fault'] = 1;
            \SwooleXRpc\Rpc::run()->set($config);
            \SwooleXRpc\Rpc::run()->ping_error($config, 4);
            //关闭连接
            socket_close($socket);
            return false;
        }
        if (!$body) {
            \SwooleXRpc\RpcRedis::run()->DECR($num_key); 
            // 这里理应关闭该连接，标记is_fault
            $config['is_fault'] = 1;
            \SwooleXRpc\Rpc::run()->set($config);
            \SwooleXRpc\Rpc::run()->ping_error($config, 4);
            $this->msg = 'connect return body Error';
            return false;
        }

        // 请求数-1
        \SwooleXRpc\RpcRedis::run()->DECR($num_key); 

        // 数据解密
        if ($this->rpc['aes_status'] == true) {
            $Currency = new \SwooleXRpc\x\Currency();
            $body = $Currency->aes_decrypt($body, $this->rpc);
            unset($Currency);
        }
        $body = json_decode($body, true);
        $this->status = $body['status'];
        $this->msg = $body['msg'];
        $this->data = $body['data'];

        if ($this->status == '200') {
            if ($this->task) {
                $this->msg = 'Task Success';
            }
            return $this->data;
        }

        return false;
    }

    // 权重获取
    private function weightConfig($list) {
        // 检测是不是刚初始化SW-X的时候
        if (empty($list[0]['ping_ms']) && empty($list[0]['is_fault'])) {
            return $list[0];
        }
        // 先删除已经不行的代码
        $yes_list = array();
        foreach ($list as $k=>$v) {
            if (isset($v['is_fault']) && $v['is_fault'] == 0) {
                $yes_list[] = $v;
            } else if (empty($v['is_fault']) && empty($v['status'])) {
                $yes_list[] = $v;
            }
        }
        // 多维数组排序
        // 评分最高，调用人数最低，延迟最低
        $score = array();
        $request_num = array();
        $ping_ms = array();
        foreach ($yes_list as $v) {
            $score[] = isset($v['score']) ? $v['score'] : 100;
            $request_num[] = isset($v['request_num']) ? $v['request_num'] : 0;
            $ping_ms[] = isset($v['ping_ms']) ? $v['ping_ms'] : 0;
        }

        $bool = array_multisort($score, SORT_DESC, $request_num, SORT_ASC, $ping_ms, SORT_ASC, $yes_list);
        return array_shift($yes_list);
    }

    // 以下为获取相关状态
    public function getMsg() {
        return $this->msg;
    }
    public function getStatus() {
        return $this->status;
    }
    public function isSuccess() {
        if ($this->status == 200) {
            return true;
        }
        return false;
    }

    
}